package environments

import (
	"context"
	"slices"
	"time"

	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"github.com/avast/retry-go/v4"
)

// Environments
// @microgen grpc
// @protobuf git.perx.ru/perxis/perxis-go/proto/environments
// @grpc-addr content.environments.Environments
type Environments interface {
	Create(ctx context.Context, env *Environment) (created *Environment, err error)
	Get(ctx context.Context, spaceId, envId string) (env *Environment, err error)
	List(ctx context.Context, spaceId string) (envs []*Environment, err error)
	Update(ctx context.Context, env *Environment) (err error)
	Delete(ctx context.Context, spaceId, envId string) (err error)
	SetAlias(ctx context.Context, spaceId, envId, alias string) (err error)
	RemoveAlias(ctx context.Context, spaceId, envId, alias string) (err error)
	Migrate(ctx context.Context, spaceId, envId string, options ...*MigrateOptions) (err error)
}

const (
	RetryDelay         = 1 * time.Second
	RetryAttempts uint = 1200
)

// IsAvailable проверяет доступность окружения.
// Если окружение недоступно, возвращает ошибку ErrEnvironmentUnavailable
func IsAvailable(ctx context.Context, svc Environments, spaceID string, environmentID string) error {
	env, err := svc.Get(ctx, spaceID, environmentID)
	if err != nil {
		return err
	}
	if env.StateInfo == nil || slices.Contains(WriteAllowedStates, env.StateInfo.State) {
		return nil
	}
	return errors.WithContext(ErrEnvironmentUnavailable, "state", env.StateInfo.State)
}

// WaitForAvailable периодически проверяет состояние окружения, пока оно не станет доступным.
// Если окружение становится доступным, функция возвращает nil.
// Если достигнуто максимальное количество попыток, функция возвращает ошибку.
//
// Через необязательные параметры можно, например, настроить задержку между попытками (по умолчанию равную RetryDelay) и максимальное количество попыток (по умолчанию равное RetryAttempts),
// но опции retry.Context и retry.RetryIf будут проигнорированы.
func WaitForAvailable(ctx context.Context, svc Environments, spaceID string, environmentID string, opts ...retry.Option) error {
	opts = slices.Concat(
		[]retry.Option{
			retry.DelayType(retry.FixedDelay),
			retry.Delay(RetryDelay),
			retry.Attempts(RetryAttempts),
		},
		opts,
		// Чтобы предотвратить изменение параметров пользователем, размещаем их в конце списка.
		[]retry.Option{
			retry.Context(ctx),
			retry.RetryIf(func(err error) bool {
				return errors.Is(err, ErrEnvironmentUnavailable)
			}),
		},
	)
	return retry.Do(
		func() error {
			return IsAvailable(ctx, svc, spaceID, environmentID)
		},
		opts...,
	)
}
